import concurrent.futures
import pickle
from copy import deepcopy as dc
from glob import glob as ls
from os import cpu_count as cpus
from os import path

import numpy as np
import scipy.stats as scist
from astropy.io import fits
from tqdm import tqdm

from . import IRIS_SG_deconvolve as isd

#There are two functions here


def ParDecon(rasfits, psfs, save=False, cdelts=None):
    '''Function acts as a wrapper around IRIS_SG_deconvolve
    Input Paramteres: 
        rasfits: The hdu to be deconvolved
        psfs: The point spread functions in a dictionary
        save: If True: Save the files with d appended
              If False: Return the deconvolved hdus
    Output:
        If save=False: Deconcolved hdu
        If save=True: 0
    '''
    hdr0=dc(rasfits[0].header)
    nlines=rasfits[0].header['NEXP']
    indices={rasfits[0].header[name]: ind+1 for ind, name in enumerate(rasfits[0].header['TDESC*'])}
    decondict={}
    hdrdict={}
    for key in indices:
        decondict[key]=np.zeros_like(rasfits[indices[key]].data)
        psfind=rasfits[0].header['TDET'+str(indices[key])]
        hdrdict[key]=dc(rasfits[indices[key]].header)
        if not cdelts:
            dy=1
        else:
            dy=cdelts[psfind]
        for j in range(0, nlines):
            decondict[key][j]=isd.IRIS_SG_deconvolve(rasfits[indices[key]].data[j], psf=psfs[psfind], fft_div=True, dy=dy)


        hdr0['TDMEAN'+str(indices[key])]=np.mean(decondict[key])
        hdr0['TDRMS'+str(indices[key])]=np.sqrt(np.sum((decondict[key]-np.mean(decondict[key]))**2)/decondict[key].size)
        hdr0['TDMEDN'+str(indices[key])]=np.median(decondict[key])
        hdr0['TDMIN'+str(indices[key])]=np.min(decondict[key])
        hdr0['TDMAX'+str(indices[key])]=np.max(decondict[key])
        hdr0['TDSKEW'+str(indices[key])]=scist.skew(decondict[key], axis=None)
        hdr0['TDKURT'+str(indices[key])]=scist.kurtosis(decondict[key], axis=None)

        flatdat=np.sort(decondict[key].flatten())
        hdr0['TDP01_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.01))]
        hdr0['TDP10_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.1))]
        hdr0['TDP25_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.25))]
        hdr0['TDP75_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.75))]
        hdr0['TDP90_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.9))]
        hdr0['TDP95_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.95))]
        hdr0['TDP98_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.98))]
        hdr0['TDP99_'+str(indices[key])]=flatdat[int(np.round(len(flatdat)*0.99))]
        del flatdat

    for ind, key in enumerate(decondict):
        if ind==0:
            dattot=decondict[key] #Needed for header stuff. (DATa TOTal)
        else:
            dattot=np.concatenate((dattot, decondict[key]), axis=2)
    hdr0['DATAMEAN']=np.mean(dattot)
    hdr0['DATARMS']=np.sqrt(np.sum((dattot-np.mean(dattot))**2)/dattot.size)
    hdr0['DATAMEDN']=np.median(dattot)
    hdr0['DATAMIN']=np.min(dattot)
    hdr0['DATAMAX']=np.max(dattot)
    hdr0['DATASKEW']=scist.skew(dattot, axis=None)
    hdr0['DATAKURT']=scist.kurtosis(dattot, axis=None)
    flatdattot=np.sort(dattot.flatten())
    hdr0['DATAP01']=flatdattot[int(np.round(len(flatdattot)*0.01))]
    hdr0['DATAP10']=flatdattot[int(np.round(len(flatdattot)*0.1))]
    hdr0['DATAP25']=flatdattot[int(np.round(len(flatdattot)*0.25))]
    hdr0['DATAP75']=flatdattot[int(np.round(len(flatdattot)*0.75))]
    hdr0['DATAP90']=flatdattot[int(np.round(len(flatdattot)*0.9))]
    hdr0['DATAP95']=flatdattot[int(np.round(len(flatdattot)*0.95))]
    hdr0['DATAP98']=flatdattot[int(np.round(len(flatdattot)*0.98))]
    hdr0['DATAP99']=flatdattot[int(np.round(len(flatdattot)*0.99))]
    del dattot, flatdattot #I imagine these are large, so delete them after they are no longer needed

    phdu=fits.PrimaryHDU(None, header=hdr0)
    hduls=[phdu]
    for key in indices:
        hduls.append(fits.ImageHDU(decondict[key], header=hdrdict[key]))
    hdul=fits.HDUList(hduls)  

    if save:
        hdul.writeto(rasfits.filename()[:-5]+'d.fits')
        return(0)
    else:
        return(hdul)


def deconvolve(ras, quiet=False, save=False, limitcores=False, cdelt=None):
    '''Function prepares input to ParDecon
    Input Paramteres: 
        ras: String, list, or astropy.io.fits.hdu.hdulist.HDUList (hdu)
             String: Path to IRIS spectrograph file
                     Path to IRIS files using wildcard (ie, /path/to/files/*fits) of same observation
             List: List of paths to spectrograph file of same observation
                   List of hdus from same observation
             hdu : An IRIS observation
        quiet: If True, suppress all print statements
        save: If True: Save the files with d appended
              If False: Return the deconvolved hdus
        limitcores: If True: use all but one core. If False use all cores. 
    Output:
        If save=False: Deconcolved hdu(s). 
        If save=True: 0
    '''


    nworkers=cpus()-int(limitcores)
    pathlistin=False
    hdulistin=False
    if type(ras)==fits.hdu.hdulist.HDUList:
        assert ras[0].header['TELESCOP']=='IRIS'
        rasfits=dc(ras)

    elif '*' in ras:
        ras=ls(rass)
        ras.sort()
        ex=fits.open(ras[0]) #example
        assert ex.header['TELESCOP']=='IRIS'
        pathlistin=True
        if not cdelt:
            cdelt={}
            for i in range(1, ex[0].header['NWIN']+1):
                if ex[0].header['TDET'+str(i)] not in cdelt:
                    cdelt[ex[0].header['TDET'+str(i)]]=ex[i].header['CDELT2']
                else:
                    assert cdelt[ex[0].header['TDET'+str(i)]]==ex[i].header['CDELT2']

    elif type(ras)==str:
        try:
            rasfits=fits.open(ras)
            if not cdelt:
                cdelt={}
                for i in range(1, rasfits[0].header['NWIN']+1):
                    if rasfits[0].header['TDET'+str(i)] not in cdelt:
                        cdelt[rasfits[0].header['TDET'+str(i)]]=rasfits[i].header['CDELT2']
                    else:
                        assert cdelt[rasfits[0].header['TDET'+str(i)]]==rasfits[i].header['CDELT2']

                    assert rasfits[0].header['TELESCOP']=='IRIS'
        except NameError:
            raise ValueError("Must supply fits file or path to fits file or * directory for one set of observations")

    elif type(ras)==list:
        if type(ras[0])==fits.hdu.hdulist.HDUList:
            assert ras[0][0].header['TELESCOP']=='IRIS'
            hdulistin=True
            if not cdelt:
                cdelt={}
                for i in range(1, ras[0][0].header['NWIN']+1):
                    if ras[0][0].header['TDET'+str(i)] not in cdelt:
                        cdelt[ras[0][0].header['TDET'+str(i)]]=ras[0][i].header['CDELT2']
                    else:
                        assert cdelt[ras[0][0].header['TDET'+str(i)]]==ras[0][i].header['CDELT2']

        else:
            try:
                ex=fits.open(ras[0])
                assert ex[0].header['TELESCOP']=='IRIS'
                pathlistin=True
                if not cdelt:
                    cdelt={}
                    for i in range(1, ex[0].header['NWIN']+1):
                        if ex[0].header['TDET'+str(i)] not in cdelt:
                            cdelt[ex[0].header['TDET'+str(i)]]=ex[i].header['CDELT2']
                        else:
                            assert cdelt[ex[0].header['TDET'+str(i)]]==ex[i].header['CDELT2']
            except NameError:
                raise ValueError("Must supply fits file or * directory for one set of observations")
    else:
        raise ValueError("Must supply fits file or * directory for one set of observations")

    toppath=path.dirname(path.realpath(__file__))
    with open(toppath+'/IRIS_SG_PSFs.pkl', 'rb') as psfpkl:
        psfsin=pickle.load(psfpkl)
    
    psfs={'FUV1':psfsin['sg_psf_1336'], 'FUV2':psfsin['sg_psf_1394'], 'NUV':psfsin['sg_psf_2796']}      

    for psfn in cdelt:
        psf=psfs[psfn]
        psfx=np.arange(0, len(psf))/6
        datx=np.arange(0, psfx[-1], cdelt[psfn])
        newpsf=np.interp(datx, psfx, psf)
        psfi=np.trapz(newpsf, datx)
        newpsf=newpsf/psfi
        psfs[psfn]=newpsf

    if pathlistin:
        with concurrent.futures.ProcessPoolExecutor(workers=nworkers) as executor:
            futures=[executor.submit(ParDecon, rasfits=fits.open(ras[i]), psfs=psfs, save=save, cdelts=cdelt) for i in range(0, len(ras))]
            for f in tqdm(concurrent.futures.as_completed(futures), total=len(rasdirec), disable=quiet):
                pass 
        out=[f for f in futures]
    elif hdulistin:
        with concurrent.futures.ProcessPoolExecutor(workers=nworkers) as executor:
            futures=[executor.submit(ParDecon, rasfits=ras[i], psfs=psfs, save=save, cdelts=cdelt) for i in range(0, len(ras))]
            for f in tqdm(concurrent.futures.as_completed(futures), total=len(rasdirec), disable=quiet):
                pass
        out=[f for f in futures]
    else:
        out=ParDecon(rasfits=ras, psfs=psfs, save=save)
    
    if not save:
        return(out)
    else:
        return(0)



import pickle
import urllib.error
import urllib.request
from datetime import datetime as dt
from glob import glob as ls
from os import path, remove

import numpy as np
from astropy.time import Time
from bs4 import BeautifulSoup
from scipy.interpolate import interp1d
from scipy.io import readsav


def iris_get_response(date=dt.strftime(dt.now(), '%Y-%m-%dT%H:%M:%S.%fZ'), version=0, response_file=None, pre_launch=False, full=False, angstrom=False):
    '''Intended to use in place of iris_get_response.pro
    Input Parameters:
        date: Time or time list. Default is now.
        version: Which version of the response file you want. Default is newest. Version numbers are 1 indexed, so the default 0 becomes -1.
        response_file: Name of the response file you want to use. Must exactly match, including extension.
        pre_launch: Not sure why this is in the original, but it is analaguous to version=2. Default is False.
        full: Full effective area structure is returned with cryptic coefficients. Default is False.
        angstrom: If True, lambda is returned in angstroms. If False, lambda is retunred in nm. Default is False.

    Notes:
        1. version, response_file, and prelaunch all perform the same function, here is their precedence,
        pre_launch>version>response_file 
        2. Code automatically checks https://hesperia.gsfc.nasa.gov/ssw/iris/response/ for new response files. If this url
        changes in the future, do a search and replace. The files are assumed to be geny IDL structs.
        3. All original comments will be preceeded by ;, as is convention in IDL
        4. Translated from iris_get_response.pro. Originally by J.P.Weulser, S.L. Freeland, and G.Chintzoglou

    History:
        2021-12-14 - A.W.Peat - Translated from IDL and added QOL improvements
    '''

    toppath=path.dirname(path.realpath(__file__))
    resppath=path.join(toppath, "responses")
    resps=ls(path.join(resppath, "*.pkl"))
    resps.sort()

    #Checks for new responses everytime it is run
    try:
        with urllib.request.urlopen("https://hesperia.gsfc.nasa.gov/ssw/iris/response/") as respurl:
            htmlsoup=BeautifulSoup(respurl, 'html.parser')
        for tags in htmlsoup.find_all('a'):
            href=tags.get('href')
            if "sra" in href and path.join(resppath, href[:-4]+'pkl') not in resps:
                print("New response file found, "+href+'.\nDownloading...')
                urllib.request.urlretrieve("https://hesperia.gsfc.nasa.gov/ssw/iris/response/"+href, "temp.geny")
                newgeny=readsav('temp.geny')
                remove(geny)
                recgeny=newgeny[list(newgeny.keys())[0]][0]
                with open(toppath+"/responses/"+href[:-4]+'pkl', "wb") as pklout:
                    pickle.dump(recgeny, pklout)

                resps=ls(toppath+"/responses/*.*") #Needs to reload responses if a new one is found
                resps.sort()
    except urllib.error.URLError:
        print("You are not connected to the internet. Cannot check for new response files.")

    #0a Opening correct file
    if pre_launch:
        response=resps[1] #0 indexing
    elif version!=0:
        if version<=0:
            print("No such version of response file. Defaulting to most recent version.")
            response=resps[-1]
        elif version<=len(resps):
            response=resps[version-1]
        else:
            print("Requested version of response file not found. Defaulting to most recent version.")
            response=resps[-1]
    elif response_file!=None:
        if toppath+"/responses/"+response_file not in resps:
            print(response_file+" not found. Using most recent file.")
            response=resps[-1]
        else:
            response="./responses/"+response_file
    else:
        response=resps[version-1]

    with open(response, "rb") as pklin:
        r=pickle.load(pklin) #Loading in the response file and calling it r

    #0b Handling keywords
    if int(r['version'])<2:
        if angstrom:
            r['lambda']=r['lambda']*10
        return(r)

    ntt=1 #ntt is always 1 since this will have a wrapper around it
    
    #1. Output structure
    if full:
        o1=r #Output 1. Is temporary
    else:
        keys=['LAMBDA', 'AREA_SG', 'NAME_SG', 'DN2PHOT_SG', 'AREA_SJI', 'NAME_SJI', 'DN2PHOT_SJI', 'COMMENT', 'VERSION', 'VERSION_DATE']
        o1={k:r[k] for k in keys}
        o1['DATE-OBS']=date

    o1['AREA_SG']=np.zeros_like(o1['AREA_SG'])
    o1['AREA_SJI']=np.zeros_like(o1['AREA_SJI'])
    o=[o1 for i in range(0, ntt)] #Output
    del o1

    #2. FUV SG Effective Areas
    # ; Rough SG spectral ranges. Setting eff.area to 0 outside of these
    lamran=[[133.1,135.9],[138.8,140.8]]
    #Not entirley sure how the coeff array is organised, but the comment on the IDL version says,
    # "; time dependent response for sz[3]=3 wavelengths". The index in Python is [0] though
    sz=r['coeffs_fuv'].shape 
    rr=np.zeros((ntt, sz[0]))
    for j in range(0, sz[0]):
        rr[:,j]=fit_iris_xput_lite(date, r['c_f_time'], r['coeffs_fuv'][j])
    #; interpolate onto lambda grid, separately for each of the two FUV CCDs
    for j in range(0, 2):
        w=np.where((r['lambda']>=lamran[j][0]) & (r['lambda']<=lamran[j][1]))
        for k in range(0, ntt):
            interp=interp1d(r['c_f_lambda'][j:j+2], rr[k, j:j+2], fill_value='extrapolate')
            #If you feel uneasy about this extrapolation, this is how iris_get_resposne.pro works implicitly
            o[k]['AREA_SG'][0,w]=interp(r['lambda'][w])


    #3. NUV SG Effective Areas
    # ; Rough SG spectral ranges. Setting eff.area to 0 outside of these
    lamran=[278.2,283.5]
    #Not entirley sure how the coeff array is organised, but the comment on the IDL version says,
    # "; time dependent response for sz[3]=3 wavelengths". The index in Python is [0] though
    sz=r['coeffs_nuv'].shape 
    rr=np.zeros((ntt, sz[0]))
    for j in range(0, sz[0]):
        rr[:,j]=fit_iris_xput_lite(date, r['c_n_time'], r['coeffs_nuv'][j])
    #; interpolate onto lambda grid
    w=np.where((r['lambda']>=lamran[0]) & (r['lambda']<=lamran[1]))
    if int(r['version'])<3:
        for k in range(0, ntt):
            interp=interp1d(r['c_n_lambda'], rr[k], fill_value='extrapolate')
            o[k]['AREA_SG'][1,w]=interp(r['lambda'][w])
    else: #I guess for version>=3, len(r['c_n_lambda'])>2
        interp=interp1d(r['c_n_lambda'], rr[k], fill_value='extrapolate', kind='quadratic')
        o[k]['AREA_SG'][1,w]=interp(r['lambda'][w])

    #4. SJI Effective Areas
    if int(r['version'])<3:
        sz=r['coeffs_sji'].shape 
        for j in range(0, sz[0]):
        # ; calculate pre-launch area from the individual elements
            pl_a=r['geom_area']
            for k in range(0, len(r['index_el_sji'][0])):
                pl_a=(pl_a*np.array([[r['elements'][r['index_el_sji'][0]][i][1]] for i in range(0, r['elements'][r['index_el_sji'][0]].shape[0])])).T
                pl_a=pl_a[:,0,:] #Because of the way I get this to work, I introduce an extra 1-length axis
            rr=fit_iris_xput_lite(time, r['c_s_time'][j], r['coeffs_sji'][j])
            for k in range(0, ntt):
                o[k]['AREA_SJI'][j]=pl_a*rr[k]

    else:
        for nuv in range(0, 2):
        # ; calculate baseline SJI area curves
            asji=r['geom_area']
            for k in range(0, len(r['index_el_sji'][nuv*2])):
                arr=np.array([[r['elements'][r['index_el_sji'][2:4, 3]][i]] for i in range(0, r['elements'][r['index_el_sji'][2:4, 3]].shape[0])])
                asji=asji*(np.array([arr[0][0][1], arr[1][0][1]])).T
                del arr
            # ; apply time dependent profile shape adjustment to FUV SJI
            if ~nuv:
                # ; FUV: apply FUV SG "slant", then normalize so that a weighted (2.4:1)
                # ;      sum at C II and Si IV gives constant response
                wei=[2.4,1] # ; typical solar ratio CII : SiIV
                wav=r['c_f_lambda']
                nwv=len(wav)
                wav=[wav[0], (wav[nwv-2]*2+wav[nwv-1])/3] # ; 2 wvlngts in nm
                # ; calculate baseline SG area for scaling purposes
                asg=r['geom_area']
                for k in range(0, len(r['index_el_sg'][nuv])):
                    asg=asg*r['elements'][r['index_el_sg']][nuv, k][1].T
                # ; SG and SJI areas at wav
                interp=interp1d(r['lambda'], asg, fill_value='extrapolate')
                asg2=interp(wav)
                asj2=np.zeros((2,2))
                for j in range(0,2):
                    interp=interp1d(r['lambda'], asji[:,j], fill_value='extrapolate')
                    asj2[:,j]=interp(wav)
                # ; calculate the normalized slant function scal, apply to asji
                for k in range(0, ntt):
                    # ; ; best-estimate slant, i.e., eff.area @ wav / baseline SG @ wav
                    interp=interp1d(o[k]['LAMBDA'], o[k]['AREA_SG'][0], fill_value='extrapolate')
                    sca2=interp(wav)/asg2
                    # ; normalize slant so that total(wei*asj2*sca2)/total(wei*asj2)=1
                    for j in range(0, 2):
                        sca2n=sca2*np.sum(wei*asj2[:,j], axis=None)/np.sum(wei*asj2[:,j]*sca2)
                        interp=interp1d(wav, sca2n, fill_value='extrapolate')
                        scaln=interp(r['lambda'])
                        o[k]['AREA_SJI'][j]=asji[:,j]*scaln
            else:
                # ; NUV: essentially same calculation as r.version=3
                for k in range(0, ntt):
                    o[k]['AREA_SJI'][2:4]=asji
            
        for j in range(0, 4):
            # ; SJI specific time dependency
            rr=fit_iris_xput_lite(date, r['c_s_time'][j], r['coeffs_sji'][j])
            for k in range(0, ntt):
                o[k]['AREA_SJI'][j]=o[k]['AREA_SJI'][j]*rr[k]

    if angstrom:
        o['lambda']=o['lambda']*10
    return(o)


def fit_iris_xput_lite(tt0, tcc0, ccc):
    '''
    Stripped down form of fit_iris_xput.pro, using only the things 
    get_iris_response.pro uses.
    I am so sorry, but I have no idea what any of these keywords are.
    The previous documentation is very cryptic. I will include ALL of their comments.

    Notes:
        1. All original comments will be preceeded by ;, as is convention in IDL
        2. Based on fit_iris_xput.pro by JPW.

    History:
        2021-12-14 - A.W.Peat - Translated from IDL
    '''
    tex=1.5 # ; exponent for transition between exp.decay intervals
    if tcc0.shape[1]!=2:
        raise RuntimeError("Incorrect number of elements in tcoef (tcco)")
    m=tcc0.size//2
    #This is crazy. Originally here they did 
    #m=size(tcc0); if m[1] ne 2; return, 0; endif; m=m[m[0]+2]/2.

    #; times in years since tcc[0,0]
    #The original is using NASA's epoch of 1979. I'm using 1970, as is standard.
    nasaLag=Time('1979-01-01T00:00:00.00Z').unix
    yr2sec=365.25*24*60*60
    tcc=(tcc0+nasaLag)/yr2sec 
    tt=Time(tt0).unix/yr2sec
    ntt=1 
    #Always going to be 1 in this instance. Script can, in theory, take more than one input.
         
    if ccc.size!=3*m:
        raise RuntimeError("Incorrect number of elements in tcoef (tcco)")

    # ; calculation of output fit
    ee=ccc[:,2] 
    a=np.zeros((ntt, 2*m)) #; base vector
    #no need to reform in python, it's 2D by default here
    for j in range(0, m):
        # ; base vector for interval of constant multiplier
        if j>0:
            ww=np.where((tt>=tcc[j, 0]) & (tt<=tcc[j, 1]))
            nww=len(ww)
        else:
            ww=np.where(tt<=tcc[j,1])
            nww=len(ww)

        if nww>0:
            a[ww, 2*j]=1
            a[ww, 2*j+1]=np.exp(ee[j]*(tt-tcc[j,0]))
        # ; base vector for interval when multiplier is linear function of time
        if j>0:
            ww=np.where((tt>tcc[j-1,1]) & (tt < tcc[j,0]))
            nww=len(ww)
            if nww>0:
                dtt=(tt-tcc[j-1,1])/(tcc[j,0]-tcc[j-1, 1])
                a[ww, 2*j]=dtt**tex
                a[ww, 2*j+1]=dtt**tex*np.exp(ee[j]*(tt-tcc[j,0]))
        if j < (m-1):
            ww=np.where((tt>tcc[j,1]) and (tt<tcc[j+1, 0]))
            nww=len(ww)
            if nww>0:
                dtt=(tt-tcc[j,1])/(tcc[j+1,0]-tcc[j,1])
                a[ww, 2*j]=1-dtt**tex
                a[ww, 2*j+1]=(1-dtt**tex)*np.exp(ee[j]*(tt-tcc[j,0]))

    cc=ccc[:,:2].flatten()
    f=a@cc
    return(f)


if __name__=="__main__":
    #When script is called directly, it just looks for new response files#
    toppath=path.dirname(path.realpath(__file__))
    resppath=path.join(toppath, "responses")
    resps=ls(path.join(resppath, "*.pkl"))
    resps.sort()
    new=False
    with urllib.request.urlopen("https://hesperia.gsfc.nasa.gov/ssw/iris/response/") as respurl:
        htmlsoup=BeautifulSoup(respurl, 'html.parser')
    for tags in htmlsoup.find_all('a'):
        href=tags.get('href')
        print(href)
        if "sra" in href and path.join(resppath, href[:-4]+'pkl') not in resps:
            new=True
            print("New response file found, "+href+'.\nDownloading...')
            urllib.request.urlretrieve("https://hesperia.gsfc.nasa.gov/ssw/iris/response/"+href, "temp.geny")
            newgeny=readsav('temp.geny')
            remove(geny)
            recgeny=newgeny[list(newgeny.keys())[0]][0]
            with open(toppath+"/responses/"+href[:-4]+'pkl', "wb") as pklout:
                pickle.dump(recgeny, pklout)
    if not new:
        print("No new response files found.")

    



import numpy as np
import scipy.fft as fft


def IRIS_SG_deconvolve(data_in, psf, 
                       iterations=10, 
                       fft_div=False,
                       dy=1):


    '''

    Graham S. Kerr
    July 2020
 
    NAME:   IRIS_SG_Deconvolve.py

    PURPOSE: Deconvolves IRIS SG data using the PSFs from Courrier et al 2018.
    
    INPUTS: data_in -- A 2D IRIS SG array [ypos, wavelength]
            psf     -- The appropriate PSF 
                       These are not currently in iris_lmsalpy, so I have just 
                       saved the IDL versions and restore them in the notebook 
                       before I call this function. 
            iterations -- The number of Richardson Lucy iterations to run through
                          Default = 10
            fft_div -- Set to use skip iterations and instead deconvolve by division
                       in Fourier Space   

    NOTES: Based on iris_sg_deconvolve.pro by Hans Courrier, but not all the functionality
           is included here yet

           There are probably more clever ways to code this in -- i'm fairly new to python. 

            To Do: Add error statements

    '''


    #Remove negative values 
    dcvim = data_in.copy()
    dcvim[dcvim<0] = 0
    data_in_zr = dcvim

    if fft_div == True:
        dcvim = FFT_conv_1D(data_in,psf,div=True)
    else:
        for ind in range(1,iterations+1):
            #print('iteration = %3d' %(ind))
            step1 = data_in_zr/(FFT_conv_1D(dcvim,psf,rev_psf=False,div=False, dy=dy))
            #print(np.nanmax(step1[265,:]))
            step2 = FFT_conv_1D(step1,psf, rev_psf=True, dy=dy)
            dcvim = dcvim * step2

    return dcvim


def FFT_conv_1D(datain, psfin, div = False, rev_psf=False, dy=1):
    

    '''   
    Notes AWP: This is not a good docstring, will fix later.

    NAME: FFT_conv_1D
    
    PURPOSE: Function to do FFT convolution in the y-direction 
              of the input data (first dimensioin). This way we 
              can pass the 1D PSF.
          
    INPUTS: datain -- a 2D data array [nominally, slit pos vs wavelength]
            psfin  -- the PSF to be applied in the y-direction
            imsize -- the dimensions of the input data array 
            psflen -- the length of the psf. 

    KEYWORDS: div -- Set to True to divide in Fourier space. 
                     Default is False, so multiply in Fourier space.
              rev_psf -- Set to reverse the 1D input PSF
              
            
    OUTPUTS: dataout -- the input data convolved with the PSF
    
    NOTES: Pretty much copied exactly from Hans Courrier's IDL version 
            in the SSW IRIS software tree, as part of iris_sg_deconvole.pro

    '''
  
    #length of input psf
    psflen = len(psfin)
    
    #dimensions of input data
    imsize = datain.shape
    
    #Get difference of image size and psf length
    ydiff = imsize[0]-psflen  
    
    #Cut the PSF if it is too long
    if ydiff < 0:
        rs = (-1*ydiff)/2
        pin=psfin[int(np.floor(rs)):int(np.ceil(rs))]
        #renormalize PSF (dx=1)
        pin = pin/np.trapz(pin, dx=dy) 
        
    #Pad the PSF if it is too short
    if ydiff > 0:
        rs = ydiff/2
        pin=np.pad(psfin, [int(np.floor(rs)), int(np.ceil(rs))])
               
    #Replicate the PSF over wavelength array also
    pin_full = np.transpose(np.tile(pin,[imsize[1],1]))
    
    #Shift PSF center to zero to center output
    pin_full = np.roll(pin_full, (int(-imsize[0]/2),0),axis=0)
        
    #Reverse the PSF if needed
    if rev_psf == True:
        pin_full = np.flip(pin_full,axis=0)        
        if psflen % 2 == 0:
            pin_full = np.roll(pin_full,(1,0),axis=0)
            
    
    #Perform the FFT
    fpsf = fft.fft(pin_full,axis=0)
    datain=datain.astype(np.float64)
    fdatain = fft.fft(datain,axis=0)
    
    #Multiply(divide) the PSF and data, and convert back to k space
    if div == False: 
        dataout = fft.ifft((fdatain*fpsf),axis=0).real
    else:
        dataout = fft.ifft((fdatain/fpsf),axis=0).real
    
    return dataout



from setuptools import setup


def readme():
    with open("README.md", "r") as f:
        return f.read()

setup(name='irispreppy',
      version="1.1.8",
      url='https://github.com/OfAaron3/irispreppy',
      author='Aaron W. Peat',
      author_email='a.peat.1@research.gla.ac.uk',
      license='MIT',
      packages=['irispreppy.psf', 'irispreppy.radcal', 'irispreppy', 'irispreppy.radcal.responses'],
      install_requires=[
          'numpy',
          'tqdm',
          'astropy',
          'scipy',
          'beautifulsoup4'
      ],
      include_package_data=True,
      zip_safe=False,
      long_description=readme(),
      long_description_content_type="text/markdown",
)
